/**
 * Copyright (c) 2014, 2015, ControlsFX
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * Neither the name of ControlsFX, any associated website, nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CONTROLSFX BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.controlsfx.validation.decoration;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import javafx.scene.control.Control;

import org.controlsfx.control.decoration.Decoration;
import org.controlsfx.validation.ValidationMessage;

/**
 * Validation decoration to combine several existing decorations into one.
 * Here is example of combining {@link GraphicValidationDecoration} and  {@link StyleClassValidationDecoration} 
 * <br> <br>
 * <img src="CompoundValidationDecoration.png" alt="Screenshot of CompoundValidationDecoration">  
 *   
 * 
 */
public class CompoundValidationDecoration extends AbstractValidationDecoration {

    private final Set<ValidationDecoration> decorators = new HashSet<>();

    /**
     * Creates an instance of validator using a collection of validators
     * @param decorators collection of validators
     */
    public CompoundValidationDecoration(Collection<ValidationDecoration> decorators) {
        this.decorators.addAll(decorators); 
    }

    /**
     * Creates an instance of validator using a set of validators
     * @param decorators set of validators
     */
    public CompoundValidationDecoration(ValidationDecoration... decorators) {
        this(Arrays.asList(decorators));
    }	

    /**
     * {@inheritDoc}
     */
	@Override
	public void applyRequiredDecoration(Control target) {
		this.decorators.stream().forEach( d -> d.applyRequiredDecoration(target)); 
	}
	
    /**
     * {@inheritDoc}
     */
	@Override
	public void applyValidationDecoration(ValidationMessage message) {
		this.decorators.stream().forEach( d -> d.applyValidationDecoration(message)); 
	}

	@Override
	protected Collection<Decoration> createValidationDecorations(ValidationMessage message) {
		return Collections.emptyList();
	}

	@Override
	protected Collection<Decoration> createRequiredDecorations(Control target) {
		return Collections.emptyList();
	}
}
